/* Simple Button Class

We provide buttons that can be triggered by hovering over virtual "hand" pointers for a specific amount of time.
The buttons have their own user-adjustable textures and rendering routine as well as an event list that can be polled by the application (event id equals a user-defined value).

*/
#pragma once

#include <GL/freeglut.h>
#include <deque>
#include <time.h>
#include <string>

// we model the button with a state machine
enum ButtonState{B_INACTIVE, B_DEFAULT, B_HOVERED, B_CLICKED};

class CircleButton
{
public:

	CircleButton(void);
	// instantiate button with user defined id (this is the button's event id), texture, position, size and color. the button's state is always initialized as inactive and invisible.
	CircleButton(std::string button_id, const char* texturefile, float xpos, float ypos, float radius, unsigned char r, unsigned char g, unsigned char b);
	~CircleButton(void);

	void drawButton();
	bool isHovered(float pointer_x, float pointer_y);
	// update the current button state (should be called each time in the main program loop)
	void updateButtonState(float pointer_x, float pointer_y);
	
	void setInactive();
	void setActive();
	void setVisible();
	// set invisible always sets button state to inactive
	void setInvisible();
	
	// set and get button appearance
	void setPosition(float xpos, float ypos);
	void setXPosition(float xpos);
	void setYPosition(float ypos);
	void setRadius(float radius);
	void setColor(unsigned char r, unsigned char g, unsigned char b);
	float getXPosition();
	float getYPosition();
	float getRadius();
	
	std::string getButtonID();
	clock_t getLastClickTime();
	
	// event handling
	static bool isEventWaiting(){return (ButtonMessageQueue.size() > 0);};
	// returns the last event and deletes it from the list
	static std::string getLastEvent();
	

private:	
	// button identification
	std::string ButtonID;

	// button appearance
	unsigned char RedChannel, GreenChannel, BlueChannel;
	float XPosition, YPosition, Radius, SqrRadius;
	GLuint BackgroundTextureID;
	GLuint SymbolTextureID;
	bool isVisible;

	// button state
	ButtonState CurrentState;
	// flag that tells that button state shall be set to inactive as soon as possible
	bool SetInactiveFlag;

	// infos about last hover and click time
	clock_t LastHoverTime;
	clock_t LastClickTime;
	// static time that must elapse before button is activated/clicked
	static long ActivationTimeOut;
	
	// list containing all the last button events
	static std::deque<std::string> ButtonMessageQueue;

	// put the current button id on the global event list 
	void notificate();

	// load the button symbol texture
	void loadTexture(const char* filename);
};

