/* 
Gesture Recognizer Class

Handles gesture spotting and recognition with the help of specific similarity measure and sliding spotting window.
Provides Loading/Saving of gesture sets (classes) and training (with navigation functionality).
The GestureRecognizer provides own event list with the following event ids:

fixed values:
DEFAULT_GESTURE
GESTURE_TRAINED
GESTURE_NOT_TRAINED

variable values:
id of gesture = zero based index of gesture set 

*/
#pragma once

#include "KinectSensor.h"
#include "Gesture.h"
#include <vector>
#include <deque>
#include <boost\thread.hpp>
#include "TimeFunctions.h"

// helper structs ****************************************************
struct RecognitionTime
{
	clock_t time;
	int GestureSetID;
};

struct RecognitionInfo
{
	NUI_SKELETON_DATA Skeleton;
	Pose pose;
	bool flag;
	int ID;
	float velocity;
};

struct RecognitionInterval
{
	int id_start;
	int id_end;
	int size;
};

struct KNNSet
{
	int ID;
	float distance;
};

struct KNNClassification
{
	int GestureSetID;
	float points;
	int id_start;
	int id_end;
	int size;
};


struct GestureSet
{
	int ID;
	std::vector<Gesture*> Gestures;
	std::vector<float> Variances;
	float MinDistance;
	Gesture MeanGesture;
	GestureType SideOfBody;
};

enum SimilarityMeasureType{EUCLIDEAN, PEARSON_CORRELATION, PEARSON_CORRELATION_SQUARED, MODIFIED_EUCLIDEAN, MODIFIED_PEARSON, DTW, MODIFIED_DTW, FM, MODIFIED_FM};

//********************************************************************


class GestureRecognizer
{

public:
	GestureRecognizer(KinectSensor* sensor);
	// instantiate with given "nr_of_neighbours" fr k-nearest-neighbours algorithm (KNN)
	GestureRecognizer(KinectSensor* sensor, bool use_knn, unsigned int nr_of_neighbours);
	~GestureRecognizer(void);

	// initialize recognition system with lots of parameters
	void initialize(float velocity_low_threshold, float velocity_high_threshold, float pose_threshold, int size_of_mean_filter, int max_length_of_gestures, int min_length_of_gestures, unsigned int recognition_window_size, int inactivity_time, int min_recognition_interval_length, float default_gesture_threshold, int resampling_size, SimilarityMeasureType measure);
	void setSimilarityMeasurement(SimilarityMeasureType measure_type, float user_threshold = -1.0f);

	// updates all deques and returns the current pose. invokes the gesture recognition steps (should be called each time in the main program loop)
	Pose update(bool spot_gesture);
	Pose update(NUI_SKELETON_DATA& skeleton, bool spot_gesture);

	// navigation through trained sets of gestures
	int setIndexToNextTrainedGestureSet();
	int setIndexToPreviousTrainedGestureSet();
	int setIndexToNextTrainedGesture();
	int setIndexToPreviousTrainedGesture();
	Gesture* getCurrentTrainedGesture();
	int removeCurrentTrainedGesture();    // returns -1 if there's no gesture to remove
	int removeCurrentTrainedGestureSet(); // returns -1 if there's no set to remove
	int getSizeOfGestureSets(){return TrainedGestureSets.size();};

	// index of the last trained gesture
	int CurrentTrainedGestureIndex;
	int CurrentTrainedGestureSetIndex;

	// infos about recognition
	float getCurrentVelocity();
	int getMaxLengthOfGestures();
	float getVelocityLowThreshold();
	float getVelocityHighThreshold();

	// go into training mode, choose between creating a new gesture set or appending (or replacing) gesture in current set
	void setTrainingMode(bool new_set = false, bool replace = false);
	bool isTrainingMode();

	// load and save
	bool saveTrainedGestures(std::string filename = "TrainedGestureSets.bin");
	bool loadTrainedGestures(std::string filename = "TrainedGestureSets.bin");

	// event handling
	bool isEventWaiting(){return (GestureMessageQueue.size() > 0);};
	// returns last gesture event and pops this element from the list
	std::string getLastEvent();


	

	// ONLY USED FOR ANALYSIS AND DATA MINING
public:

	int CURRENT_SIZE_OF_MOVING_WINDOW;
	float CURRENT_VELOCITY;
	int IGNORE_INDEX;
	int IGNORE_SET;

	//************** statistical analysis ***************
	// print out trained gestures for statistical aggregation
	void stat_printGestures();
	void stat_printValidateGestures(std::vector<std::vector<Gesture*>> gestures_to_validate, int minimum_neighbours, int maximum_neighbours, SimilarityMeasureType measure_type, float user_threshold=-1, const char* filename="DEFAULT.txt");
	void stat_printCalculationOfMeanGestures();
	void stat_recognizeGesture(Gesture& gesture, KNNClassification& winner){recognizeGesture(gesture, winner);};





private:
	// spots a gesture in the poses stream with the help of the sliding window containing time indices of "interesting" poses
	void spotGesture(std::deque<int> moving_window);
	// finds the boundaries of temporarily gestures that shall be checked in the sliding window
	std::vector<RecognitionInterval> findRecognitionIntervals(std::deque<int> &moving_window);
	// prepares one temporarily gesture with time indices from sliding window and checks out gesture style
	KNNClassification recognizeInterval(RecognitionInterval& interval);
	// checks time indices in sliding window and pops invalid elements
	void updateMovingWindow();
	
	// classifies a temporarily gesture with the systems similarity measurement
	void recognizeGesture(Gesture& gesture, KNNClassification& winner);
	// match gesture against default position
	bool isDefaultGesture(Gesture& gest);
	// wrapper method for gesture distance calculation
	float getGestureDistance(Gesture &gest1, Gesture &gest2, int gesture_set_index);

	// calculates pose distances of recgnition info to first and last poses of all trainied data
	bool checkPose(RecognitionInfo &info);
	// finds time boundaries of training sample
	void findTrainingInterval();
	// set all the poses flags to given value
	void setAllPosesFlags(bool value);

	// statistical helper methods
	void setMinDistances();
	void setVariances(int index);
	void setAllVariances();
	void calculateMeanGesture(int gesture_set_index);
	void calculateAllMeanGestures();
	void standardizeAllTrainedGestures();
	


	// returns an id for a recognition info structure
	int getNextRecInfoID();
	// returns index of recognition info in the recognition list (= list of the last x poses) for given id
	int getIndexFromID(int id);



//*	<- remove first slash when not using data mining functions 
public:
//*/
	// distance calculations for particular similarity measures
	float getDTWGestureDistance(Gesture &gest1, Gesture &gest2);
	float getModifiedDTWGestureDistance(Gesture &gest1, Gesture &gest2);
	float getFMGestureDistance(Gesture& gest1, Gesture& gest2);
	float getModifiedFMGestureDistance(Gesture& gest1, int gesture_set_index, int exclude_gest=-1);
	float getPearsonGestureDistance(Gesture& gest1, Gesture& gest2, int gesture_set_index);
	float getModifiedEuclideanGestureDistance(Gesture& gest1, Gesture& gest2);

	// list for gesture events
	std::deque<std::string> GestureMessageQueue;
	// current similarity measure
	SimilarityMeasureType SimilarityMeasure;
	// radius of filter for poses filtering in the update phase
	int SizeOfMeanFilter;
	// if gesture is too long we resample it down to ResamplingSize
	unsigned int ResamplingSize;
	// Moving window for interval recognition
	std::deque<int> MovingWindow;
	// ID for recognition infos
	int RecInfoID;
	// which method should we use? k-nearest-neighbour or distance to mean gesture
	bool UseKNNAlgorithm;
	// in case of knn, how many neighbour should we use?
	unsigned int NrOfNeighbours;
	// pointer to kinect sensor
	KinectSensor* NuiSensor;

	// thresholds
	float VelocityHighThreshold;
	float VelocityLowThreshold;
	float RecognitionThreshold;
	float PoseThreshold;

	//training mode flags
	bool TrainingMode;
	bool ReplaceGesture;
	bool NewGestureSet;

	// contains the last poses for mean filtering
	std::deque<Pose> LastPoses;

	// default gesture (hard-coded)
	Gesture DefaultGesture;

	// the last pose
	Pose LastPose;

	// distance threshold to accept default pose
	float DefaultGestureDistanceThreshold;

	// time threshold for inactivity, used in training mode. this is the number of contigous frames where the velocity value is below the search threshold
	int InactivityTime;
	// minimum interval length between two marked poses 
	int MinimumRecognitionIntervalLength;

	// Vector containing all the already trained gestures
	std::vector<GestureSet*> TrainedGestureSets;

	// list containing all the smoothed poses. over this list we build up the sliding window.
	std::deque<RecognitionInfo> RecognitionList;
	// maximum number of marked poses in the sliding window
	unsigned int RecognitionWindowSize;
	// the size of the recognition list
	int MaxLengthOfGestures;
	// minimum size that gestures must have to accept them in the sliding window
	int MinLengthOfGestures;

	// info about last recognition event
	RecognitionTime LastRecognitionTime;

	// mutex
	boost::mutex MUTEX;
	// ensure that only one recognition thread is running
	bool IS_RECOGNITION_THREAD_RUNNING;
};

