#include "HandRecognizer.h"
#include <iostream>
#include <math.h>
#include <fstream>
#include "TimeFunctions.h"


#define PI (3.14159265f)

// global event list 
std::deque<std::string> HandRecognizer::HandMessageQueue;


HandRecognizer::HandRecognizer(KinectSensor* sensor, HandType which_hand)
{
	NuiSensor = sensor;

	HandFlags.assign(3, false);
	
	State = H_OPEN;
	
	LastHandImage.width = 0;
	LastHandImage.height = 0;
	LastHandImage.imagedata = new BYTE[0];

	HandIsInFront = false;
	
	DepthThreshold = 110;
	SizeOfMaximaNeighbourHood = 8;
	SizeOfContourFilter = 6;
	ContourGradient = 0.6f;

	SideOfBody = which_hand;

	if(SideOfBody == H_LEFT)
	{
		WristSkeletonIndex = 6;
		ElbowSkeletonIndex = 5;
	}
	else
	{
		WristSkeletonIndex = 10;
		ElbowSkeletonIndex = 9;
	}

	
}


HandRecognizer::~HandRecognizer(void)
{
}

void HandRecognizer::setBoundingBox(const USHORT* depth_map, int depth, int hand_x, int hand_y, float angle_val, int &left_boundary, int &right_boundary, int &top_boundary, int &bottom_boundary)
{
	bool inside_bounding_box;
	int search_x; 
	int search_y;

	// we try to find the left boundary
	for (int x = 0; x > -100; x--)
	{
		inside_bounding_box = false;
		for (int y = 0; y < 200; y++)
		{
			// calculate the rotated and translated point
			search_x = hand_x + (int) (cos(angle_val) * x - sin(angle_val) * y + 0.5f);
			search_y = hand_y - (int) (sin(angle_val) * x + cos(angle_val) * y + 0.5f); // we pay attention to the depth map coordinates with (0,0) in the upper left corner (hence we mirror the y-axis)


			// if we find a valid depth value difference along the search line we break and set the boundary
			if ( (search_x > 0) && (search_x < 640) && (search_y > 0) && (search_y < 480))
			{
				if ( abs(depth_map[search_y*640 + search_x] - depth) < (int) DepthThreshold )
				{
					inside_bounding_box = true;
					left_boundary = x;
					break;
				}
			}
		}
		if (!inside_bounding_box)
			break;
	}

	// we try to find the right boundary
	for (int x = 0; x < 100; x++)
	{
		inside_bounding_box = false;
		for (int y = 0; y < 200; y++)
		{
			// calculate the rotated and translated point
			search_x = hand_x + (int) (cos(angle_val) * x - sin(angle_val) * y + 0.5f);
			search_y = hand_y - (int) (sin(angle_val) * x + cos(angle_val) * y + 0.5f);


			// if we find a valid depth value difference along the search line we break and set the boundary
			if ( (search_x > 0) && (search_x < 640) && (search_y > 0) && (search_y < 480))
			{
				if ( abs(depth_map[search_y*640 + search_x] - depth) < (int) DepthThreshold )
				{
					inside_bounding_box = true;
					right_boundary = x;
					break;
				}	
			}
		}
		if (!inside_bounding_box)
			break;
	}
			

	// we try to find the top boundary
	for (int y = 0; y < 200; y++)
	{
		inside_bounding_box = false;
		for (int x = left_boundary; x < right_boundary; x++)
		{
			// calculate the rotated and translated point
			search_x = hand_x + (int) (cos(angle_val) * x - sin(angle_val) * y + 0.5f);
			search_y = hand_y - (int) (sin(angle_val) * x + cos(angle_val) * y + 0.5f);


			// if we find a valid depth value difference along the search line we break and set the boundary
			if ( (search_x > 0) && (search_x < 640) && (search_y > 0) && (search_y < 480))
			{
				if ( abs(depth_map[search_y*640 + search_x] - depth) < (int) DepthThreshold )
				{
					inside_bounding_box = true;
					top_boundary = y;
					break;
				}	
			}
		}
		if (!inside_bounding_box)
			break;
	}

	//if(HandType == "RIGHT") std::cout<<"\rleft: "<<left_boundary<<"    right: "<<right_boundary<<"   top: "<<top_boundary<<"            ";
}


void HandRecognizer::update()
{
	clock_t start = TimeFunctions::getTime();

	// get the skeleton
	NUI_SKELETON_DATA skeleton = NuiSensor->getCurrentSkeletonData();

	// we need to find the hand position in the depth map as well as its depth
	long hand_x, hand_y;
	unsigned short hand_depth;
	NuiTransformSkeletonToDepthImage(skeleton.SkeletonPositions[WristSkeletonIndex], &hand_x, &hand_y, &hand_depth, NUI_IMAGE_RESOLUTION_640x480);
	// discard player info
	int depth = (int) (hand_depth >> 3);

	
	// then we check if the hand is in front of the body, else we do not start any recognition 
	for (int i = 0; i < 12; i++)
	{
		// we ignore the elbow, wrist and hand positions (9 10 11  and    5 6 7)
		if ( ((i >= 9) && (i <= 11)) || ((i >= 5) && (i <= 7))  )
		{
			continue;
		}
		else
		{
			long test_x, test_y;
			unsigned short test_depth;
			NuiTransformSkeletonToDepthImage(skeleton.SkeletonPositions[i], &test_x, &test_y, &test_depth, NUI_IMAGE_RESOLUTION_640x480);
			test_depth = test_depth >> 3;
			if ( (test_depth) < depth )
			{
				HandIsInFront = false;
				return;
			}
		}
	}

	HandIsInFront = true;

	// find the elbow
	long elbow_x, elbow_y;
	unsigned short elbow_depth;
	NuiTransformSkeletonToDepthImage(skeleton.SkeletonPositions[ElbowSkeletonIndex], &elbow_x, &elbow_y, &elbow_depth, NUI_IMAGE_RESOLUTION_640x480);

	// get the angle between wrist and elbow with respect to the y-axis
	float angle_val = 0;
	float angle_sign;

	// if elbow is on the right our sign shall be "+"
	if (elbow_x - hand_x > 0)
	{
		angle_sign = 1.0f;
	}
	else
	{
		angle_sign = -1.0f;
	}

	angle_val = angle_sign * atan2((float) abs(elbow_x - hand_x), (float) abs(elbow_y - hand_y));
	
	// if the elbow is above the wrist we have an angle greater 90 degrees
	if (elbow_y < hand_y)
	{
		angle_val = PI - angle_val;
	}

	// now we try to find a bounding box in the rotated coordinate system with hand position as origin	
	const USHORT* depth_map = NuiSensor->getDepthMap();

	int left_boundary = 0;
	int right_boundary = 0;
	int top_boundary = 0;
	int bottom_boundary = 0;

	setBoundingBox(depth_map, depth, hand_x, hand_y, angle_val, left_boundary, right_boundary, top_boundary, bottom_boundary);

	// if we do not find a bounding box we return
	if(left_boundary == 0 && right_boundary == 0)
	{
		return;
	}

	// delete hand pixel distances
	ContourPixelsDistances.clear();
	// from the wrist we generate search lines in angular directions over 4 octants to find the hand contour
	int search_x;
	int search_y;
	ContourPixelsDistancesMean = 0;
	int sum_of_sqr_dist = 0;
	float angle_step_size = 0.5f;
	std::vector<math::Vec2f> contour_pixels;

	// OCTANT I
	for (float alpha = 0; alpha < 45 ; alpha += angle_step_size)
	{
		// we step along the search line from the right boundary to the center. in this direction we find the hand contour much faster than starting from the center
		int x = right_boundary;
		int y = (int) (tan(alpha * PI/180.0f) * x + 0.5f);

		while (x > 0)
		{
			// calculate the rotated and translated point
			search_x = hand_x + (int) (cos(angle_val) * x - sin(angle_val) * y + 0.5f);
			search_y = hand_y - (int) (sin(angle_val) * x + cos(angle_val) * y + 0.5f); // we pay attention to the depth map coordinates with (0,0) in the upper left corner (hence we mirror the y-axis)

			// if we find a valid depth value difference along the search line we break. furthermore we calculate the length of the scan line and push it on the distances stack
			if ( (search_x > 0) && (search_x < 640) && (search_y > 0) && (search_y < 480))
			{
				if ( abs(depth_map[search_y*640 + search_x] - depth) < (int) DepthThreshold )
				{
					int sqr_dist = x*x + y*y;
					sum_of_sqr_dist += sqr_dist;
					ContourPixelsDistances.push_back(sqrtf((float) sqr_dist));
					ContourPixelsDistancesMean += ContourPixelsDistances.back();
					contour_pixels.push_back(math::Vec2f((float) x,(float) y));
					break;
				}	
			}
			// for the next step we decrement x and calculate the corresponding y
			x--;
			y = (int) (tan(alpha * PI/180.0f) * x + 0.5f);
		}
	}

	// OCTANT II
	for (float alpha = 45; alpha > 0 ; alpha -= angle_step_size)
	{
		// we step along the search line from the top boundary to the center. in this direction we find the hand contour much faster than starting from the center
		int y = top_boundary;
		int x = (int) (tan(alpha * PI/180.0f) * y + 0.5f);
		
		while (y > 0)
		{	
			search_x = hand_x + (int) (cos(angle_val) * x - sin(angle_val) * y + 0.5f);
			search_y = hand_y - (int) (sin(angle_val) * x + cos(angle_val) * y + 0.5f); // we pay attention to the depth map coordinates with (0,0) in the upper left corner (hence we mirror the y-axis)

			// if we do find a valid depth value difference along the search line we break. furthermore we calculate the length of the scan line and push it on the distances stack
			if ( (search_x > 0) && (search_x < 640) && (search_y > 0) && (search_y < 480))
			{
				if ( abs(depth_map[search_y*640 + search_x] - depth) < (int) DepthThreshold )
				{
					int sqr_dist = x*x + y*y;
					sum_of_sqr_dist += sqr_dist;
					ContourPixelsDistances.push_back(sqrtf((float) sqr_dist));
					ContourPixelsDistancesMean += ContourPixelsDistances.back();
					contour_pixels.push_back(math::Vec2f((float) x, (float) y));
					break;
				}	
			}
			

			// for the next step we decrement y and calculate the corresponding x
			y--;
			x = (int) (tan(alpha * PI/180.0f) * y + 0.5f);
		}
	}

	// OCTANT III
	for (float alpha = 0; alpha < 45 ; alpha += angle_step_size)
	{
		// we step along the search line from the left boundary to the center. in this direction we find the hand contour much faster than starting from the center
		int y = top_boundary;
		int x = - (int) (tan(alpha * PI/180.0f) * y + 0.5f);

		while (y > 0)
		{
			// calculate the rotated and translated point
			search_x = hand_x + (int) (cos(angle_val) * x - sin(angle_val) * y + 0.5f);
			search_y = hand_y - (int) (sin(angle_val) * x + cos(angle_val) * y + 0.5f); // we pay attention to the depth map coordinates with (0,0) in the upper left corner (hence we mirror the y-axis)

			// if we do find a valid depth value difference along the search line we break. furthermore we calculate the length of the scan line and push it on the distances stack
			if ( (search_x > 0) && (search_x < 640) && (search_y > 0) && (search_y < 480))
			{
				if ( abs(depth_map[search_y*640 + search_x] - depth) < (int) DepthThreshold )
				{
					int sqr_dist = x*x + y*y;
					sum_of_sqr_dist += sqr_dist;
					ContourPixelsDistances.push_back(sqrtf((float) sqr_dist));
					ContourPixelsDistancesMean += ContourPixelsDistances.back();
					contour_pixels.push_back(math::Vec2f((float) x, (float) y));
					break;
				}	
			}
			// for the next step we decrement y and calculate the corresponding x
			y--;
			x = - (int) (tan(alpha * PI/180.0f) * y + 0.5f);
		}
	}

	// OCTANT IV
	for (float alpha = 45; alpha >= 0 ; alpha -= angle_step_size)
	{
		// we step along the search line from the boundary to the center. in this direction we find the hand contour much faster than starting from the center
		int x = left_boundary;
		int y = (int) (tan(alpha * PI/180.0f) * (-x) + 0.5f);
		while (x < 0)
		{
			// calculate the rotated and translated point
			search_x = hand_x + (int) (cos(angle_val) * x - sin(angle_val) * y + 0.5f);
			search_y = hand_y - (int) (sin(angle_val) * x + cos(angle_val) * y + 0.5f); // we pay attention to the depth map coordinates with (0,0) in the upper left corner (hence we mirror the y-axis)

			// if we do find a valid depth value difference along the search line we break. furthermore we calculate the length of the scan line and push it on the distances stack
			if ( (search_x > 0) && (search_x < 640) && (search_y > 0) && (search_y < 480))
			{
				if ( abs(depth_map[search_y*640 + search_x] - depth) < (int) DepthThreshold )
				{
					int sqr_dist = x*x + y*y;
					sum_of_sqr_dist += sqr_dist;
					ContourPixelsDistances.push_back(sqrtf((float) sqr_dist));
					ContourPixelsDistancesMean += ContourPixelsDistances.back();
					contour_pixels.push_back(math::Vec2f((float) x, (float) y));
					break;
				}			
			}	
			// for the next step we increment x and calculate the corresponding y
			x++;
			y = (int) (tan(alpha * PI/180.0f) * (-x) + 0.5f);
		}
	}

	// we smooth the distance data with mean filter of radius SizeOfContourFilter
	unsigned int size = ContourPixelsDistances.size();
	if ( size > 2 * SizeOfContourFilter)
	{
		std::vector<float> temp = ContourPixelsDistances;
		float moving_sum = 0.0f;
		float scale = 1.0f / ( (float) (2 * SizeOfContourFilter + 1) );

		// we fill the moving sum 
		for (unsigned int i = 0; i < 2 * SizeOfContourFilter + 1; i++)
		{
			moving_sum += temp[i];
		}
		// the first "filter_size" values should be all the same
		for (unsigned int i = 0; i < SizeOfContourFilter; i++)
		{
			ContourPixelsDistances[i] = scale * moving_sum;
		}
		// we iterate through the data
		for (unsigned int i = SizeOfContourFilter; i < size - SizeOfContourFilter - 1; i++)
		{
			ContourPixelsDistances[i] = scale * moving_sum;
			moving_sum += temp[i+SizeOfContourFilter+1]; 
			moving_sum -= temp[i-SizeOfContourFilter]; 
		}
		// the last "filter_size" entries also shall contain the same value
		for (unsigned int i = size-1; i >= size - SizeOfContourFilter - 1; i--)
		{
			ContourPixelsDistances[i] = scale * moving_sum;
		}
	}
	// we calculate the mean of the distance data
	float N = (float) ContourPixelsDistances.size();
	ContourPixelsDistancesMean /= N;

	MaximaOfContourPixelsDistances.clear();

	// if we have not enough entries in the distances vector we return
	if(ContourPixelsDistances.size() < 2 * SizeOfMaximaNeighbourHood)
	{
		return;
	}

	// we search for local maxima within the interval [-neighbourhood, neighbourhood]
	for (unsigned int i = SizeOfMaximaNeighbourHood; i < size - SizeOfMaximaNeighbourHood; i++)
	{
		bool maximum = true;

		// we only check for a local maximum if the distance is greater than the mean
		if (ContourPixelsDistances[i] > ContourPixelsDistancesMean)
		{
			// we check the left side
			for (unsigned int j = 0; j < SizeOfMaximaNeighbourHood; j++)
			{
				maximum = maximum && (ContourPixelsDistances[i-j-1] <= ContourPixelsDistances[i-j]);
			} 
			// we check the right side
			for (unsigned int j = 0; j < SizeOfMaximaNeighbourHood; j++)
			{
				maximum = maximum && (ContourPixelsDistances[i+j+1] <= ContourPixelsDistances[i+j]);
			} 
			// if we found a maximum we check the shape of the "hill" in the histogramm to decide whether we found a finger or not
			if (maximum)
			{
				float gradient_left = (ContourPixelsDistances[i] - ContourPixelsDistances[i-SizeOfMaximaNeighbourHood]) / ((float) SizeOfMaximaNeighbourHood);
				float gradient_right = (ContourPixelsDistances[i] - ContourPixelsDistances[i+SizeOfMaximaNeighbourHood]) / ((float) SizeOfMaximaNeighbourHood);
				
				if ( (gradient_left + gradient_right) > ContourGradient)
				{
					MaximaOfContourPixelsDistances.push_back(i);
				}
			}
		}
	}

	// we decide whether hand is open or not by validating the number of the local maxima
	bool hand_closed = (MaximaOfContourPixelsDistances.size() < 2); // <-- sometimes we get a maximum on a closed hand


	// let's generate a pixel image
	int width = right_boundary - left_boundary + 1;
	int height = top_boundary - bottom_boundary + 1;
	unsigned char val = 0;

	// free the memory from last hand image
	delete[] LastHandImage.imagedata;

	// generate new hand image
	BYTE* imagedata = new BYTE[width*height*4];

	int index = 0;

	for (int y = top_boundary; y > bottom_boundary - 1; y--)
	{
		for (int x = left_boundary; x < right_boundary + 1; x++)
		{
			val = 0;

			int search_x = hand_x + (int) (cos(angle_val) * x - sin(angle_val) * y + 0.5f);
			int search_y = hand_y - (int) (sin(angle_val) * x + cos(angle_val) * y + 0.5f);

			if ( (search_x > 0) && (search_x < 640) && (search_y > 0) && (search_y < 480))
			{
				if ( abs(depth_map[search_y*640 + search_x] - depth) < (int) DepthThreshold )
				{
					val = 180;
				}	
			}
			imagedata[index*4] = val;
			imagedata[index*4+1] = val;
			imagedata[index*4+2] = val;
			imagedata[index*4+3] = 0;	
			index++;
		}
	}


	//// show contour pixels and mean of contour for visual analysis
	//int mean_x = 0;
	//int mean_y = 0;
	//int count = 0;
	//for(unsigned int i = 0; i < contour_pixels.size(); i++)
	//{
	//	int x = contour_pixels[i].x;
	//	int y = contour_pixels[i].y;
	//	x -= left_boundary;
	//	y = top_boundary - y;

	//	if((y >= 0) && (y < width) && (x >= 0) && (x < height))
	//	{
	//		mean_x += x;
	//		mean_y += y;
	//		count++;
	//		imagedata[(width*y + x)*4] = 255;
	//		imagedata[(width*y + x)*4+1] = 255;
	//		imagedata[(width*y + x)*4+2] = 255;
	//		imagedata[(width*y + x)*4+3] = 0;
	//	}
	//}

	//// show mean
	//if(count > 0)
	//{
	//	mean_x = (int) (((float) mean_x) /((float) count) + 0.5f);
	//	mean_y = (int) (((float) mean_y) /((float) count) + 0.5f);

	//	imagedata[(width*mean_y + mean_x)*4] = 255;
	//	imagedata[(width*mean_y + mean_x)*4+1] = 0;
	//	imagedata[(width*mean_y + mean_x)*4+2] = 255;
	//	imagedata[(width*mean_y + mean_x)*4+3] = 0;
	//}


	// update deque
	HandFlags.pop_front();
	HandFlags.push_back(hand_closed);
	
	// set hand image
	LastHandImage.width = width;
	LastHandImage.height = height;
	LastHandImage.imagedata = imagedata;

	// update state
	updateState();


	//std::cout<<TimeFunctions::getTime() - start<<"\n";

	// if we want to capture the parameters for statistical reasons
	if (stat_captureParams)
	{
		// capture closed hand frames
		if (stat_HandClosed)
		{
			stat_ClosedHandParams.push_back(math::Vec3f((float) depth, (float) (TimeFunctions::getTime()-start), (float) hand_closed));
		}
		// capture open hand frames
		else
		{
			stat_OpenHandParams.push_back(math::Vec3f((float) depth, (float) (TimeFunctions::getTime()-start), (float) !hand_closed));
		}
	}
}


void HandRecognizer::updateState()
{
	bool closed;
	std::string which_hand = (SideOfBody == H_LEFT)?("LEFT"):("RIGHT"); 
	
	switch (State)
	{
		case H_OPEN:
			closed = 1;
			// if all hand flags are 1 we switch the state to closed and write event to list
			for (unsigned int i = 0; i < HandFlags.size(); i++)
			{
				closed = closed && HandFlags[i];
			}
			if (closed)
			{
				HandMessageQueue.push_back(which_hand+"_HAND_CLOSED");
				State = H_CLOSED;
			}
			break;

		case H_CLOSED:
			closed = 0;
			// if all left hand flags are 0 we switch the state to open and write event to list
			for (unsigned int i = 0; i < HandFlags.size(); i++)
			{
				closed = closed || HandFlags[i];
			}
			if (!closed)
			{
				HandMessageQueue.push_back(which_hand+"_HAND_OPENED");
				State = H_OPEN;
			}
			break;
	}
}


HandImage* HandRecognizer::getLastHandImageData()
{
	return &LastHandImage;
}

HandState HandRecognizer::getHandState()
{
	return State;
}

std::string HandRecognizer::getLastEvent()
{
	std::string ev = HandMessageQueue.front();
	HandMessageQueue.pop_front();
	return ev;
}






//statistical analysis ****************************************************************************************************

void HandRecognizer::stat_setCapture(bool trigger, bool hand_closed)
{
	stat_captureParams = trigger;
	stat_HandClosed = hand_closed;
}

void HandRecognizer::stat_printStats()
{
	std::ofstream file1 ("hand_closed.txt", std::ios::trunc);
	if (file1.is_open())
	{
		for (unsigned int i=0; i < stat_ClosedHandParams.size(); i++)
		{
			file1<<stat_ClosedHandParams[i].x<<"\t"<<stat_ClosedHandParams[i].y<<"\t"<<stat_ClosedHandParams[i].z<<"\t"<<"\n";
		}
		
		file1.close();
	}
	else std::cout << "Unable to open file";


	std::ofstream file2 ("hand_open.txt", std::ios::trunc);
	if (file2.is_open())
	{
		for (unsigned int i=0; i < stat_OpenHandParams.size(); i++)
		{
			file2<<stat_OpenHandParams[i].x<<"\t"<<stat_OpenHandParams[i].y<<"\t"<<stat_OpenHandParams[i].z<<"\t"<<"\n";
		}
		
		file2.close();
	}
	else std::cout << "Unable to open file";

}