#include "RandomizeBrush.h"
#include "SOIL.h"

#define PI (3.14159f)

RandomizeBrush::RandomizeBrush(KinectPointer* right_hand, GridMesh* mesh)
{
	RightHandPointer = right_hand;
	BrushRadius = 4.0f;
	BrushStrength = 0.5f;
	Mesh = mesh;
	showBrush = false;
	Scale = 1.0f;

	TextureID = SOIL_load_OGL_texture("../images/randomize_brush.png", SOIL_LOAD_AUTO, SOIL_CREATE_NEW_ID, SOIL_FLAG_MIPMAPS);
}



RandomizeBrush::~RandomizeBrush(void)
{
}


void RandomizeBrush::updateLeftHandInput(Pose &pose)
{
	setRadiusAndStrength(pose);
	LastPose = Pose(pose);
}

void RandomizeBrush::updateRightHandInput(Pose &pose)
{
	processMesh();
	LastPose = Pose(pose);
}

void RandomizeBrush::updateBothHandInput(Pose &pose)
{
	setRadiusAndStrength(pose);
	processMesh();
	LastPose = Pose(pose);
}

void RandomizeBrush::update(Pose &pose)
{
	showBrush = (RightHandPointer->isActive())?(Mesh->ScreenToLocalCoords(RightHandPointer->getWindowCoordinates(), BrushCoordinates)):(false);
	LastPose = Pose(pose);
}

void RandomizeBrush::render()
{
	// render brush
	if (showBrush)
	{
		GLUquadric* quadric = gluNewQuadric();
		glPushMatrix();
		glTranslatef(BrushCoordinates.x, BrushCoordinates.y + .1f, BrushCoordinates.z);

		// we visualize sign of strength in different colors
		if (BrushStrength < 0.0f)
		{
			glColor3f(1.0f - abs(BrushStrength), 1.0f - abs(BrushStrength), 1.0f);
		}
		else
		{
			glColor3f(1.f, 1.0f - BrushStrength, 1.0f - BrushStrength);
		}

		gluSphere(quadric,0.2,10,10);
		glRotatef(-90, 1.0f, 0.0f, 0.0f);
		
		gluDisk(quadric, BrushRadius - abs(BrushStrength)*0.5f - 0.2f, BrushRadius + abs(BrushStrength)*0.5f + 0.2f, 20, 1);


		gluDeleteQuadric(quadric);
		glPopMatrix();
	}
}

void RandomizeBrush::setRadiusAndStrength(Pose& pose)
{
	float diff_x = pose.LeftHandPosition.x - LastPose.LeftHandPosition.x;
	float diff_y = pose.LeftHandPosition.y - LastPose.LeftHandPosition.y;

	if ( abs(diff_x) > abs(diff_y))
	{
		BrushRadius += 50 * diff_x;
		BrushRadius = max(0.0f, BrushRadius);
	}
	else 
	{
		BrushStrength += diff_y;
		BrushStrength = max(-1.0f, BrushStrength);
		BrushStrength = min(1.0f, BrushStrength);
	}
}

void RandomizeBrush::processMesh()
{
	showBrush = Mesh->ScreenToLocalCoords(RightHandPointer->getWindowCoordinates(), BrushCoordinates);
	// if brush is inside grid area
	if(showBrush)
	{
		int radius = (int) BrushRadius;
		int brush_x = (int) BrushCoordinates.x;
		int brush_z = (int) BrushCoordinates.z;

	
		// iterate over all the grid points in the radius area
		for (int x = brush_x -radius; x <= brush_x + radius; x++)
		{
			if ( (x >= 0) && (x <= Mesh->getWidth()) )
			{
				for (int z = brush_z - radius; z <= brush_z + radius; z++)
				{
					if ( (z >= 0) && (z <= Mesh->getLength()) )
					{
						GridPoint* p = Mesh->getGridPoint(x, z); 
						// we check if grid point is inside the radius
						float v1 = p->coords.x - BrushCoordinates.x;
						float v2 = p->coords.z - BrushCoordinates.z;
						float dist = sqrt(v1 * v1 + v2 * v2);
						if( dist < radius )
						{
							// for every grid point we add/subtract a bit of random height 
							float val; 
							RandomNumberGenerator.uniform(val);

							p->coords.y +=  val * Scale * BrushStrength * pow(2.71828f, - 4*(dist/BrushRadius)*(dist/BrushRadius));

							// put the grid point into the normal recalculation list
							Mesh->recalculateGridPointNormal(p);
						}
					}	
				}
			}
		}
		// recalculate the involved grid point normals
		Mesh->executeNormalsRecalculation();
	}
}