#include "Rubber2D.h"
#include "SOIL.h"


Rubber2D::Rubber2D(KinectPointer* right_hand, Manipulator2D* manipulator, std::vector<Stroke>* strokes, int* selected_stroke_id)
{
	RightHandPointer = right_hand;
	Manipulator = manipulator;
	Strokes = strokes;
	SelectedStrokeID = selected_stroke_id;
	RubberRadius = 10.0f;
	TextureID = SOIL_load_OGL_texture("../images/rubbertool.png", SOIL_LOAD_AUTO, SOIL_CREATE_NEW_ID, SOIL_FLAG_MIPMAPS);
}
Rubber2D::~Rubber2D()
{
}

void Rubber2D::updateLeftHandInput(Pose &pose)
{
	setRubberSize(pose);
	// update pose
	LastPose = Pose(pose);
}


void Rubber2D::updateRightHandInput(Pose &pose)
{
	rub();
	// update pose
	LastPose = Pose(pose);
}	

void Rubber2D::updateBothHandInput(Pose &pose)
{
	setRubberSize(pose);
	rub();
	// update pose
	LastPose = Pose(pose);
}
	

void Rubber2D::update(Pose &pose)
{
	// update pose
	LastPose = Pose(pose);
}	

void Rubber2D::render()
{
	if(RightHandPointer->isActive())
	{
		math::Vec2f coords =  RightHandPointer->getWindowCoordinates();

		glPushMatrix();
		glTranslatef(coords.x, coords.y, 0.0f);
		GLUquadric *quadric = gluNewQuadric();
		glColor3f(0.0f, 0.0f, 0.0f);
		gluDisk(quadric, RubberRadius - 0.05*RubberRadius, RubberRadius, 16, 1);			
		glPopMatrix();

		// free memory
		gluDeleteQuadric(quadric);
	}
}

void Rubber2D::rub()
{
	if(*SelectedStrokeID != -1)
	{
		// if we have a valid stroke id we try to get the corresponding stroke in the list
		for(unsigned int i = 0; i < Strokes->size(); i++)
		{
			if( (*Strokes)[i].ID == (*SelectedStrokeID) )
			{
				// save an undo information
				RubberInfo undo_info;
				// we copy the old stroke
				undo_info.OldStroke = (*Strokes)[i];

				// rubber tool in world coordinates
				math::Vec2f rubber_world_coords = RightHandPointer->getWindowCoordinates();
				// get the local coordinates of the pointer...
				math::Vec2f coords = Manipulator->WorldToLocalCoords(rubber_world_coords);
				//... and the radius
				float rubber_radius = (coords - Manipulator->WorldToLocalCoords(rubber_world_coords + math::Vec2f(RubberRadius,0.0f))).getLength();

				// if stroke contains only one element, we delete it
				if((*Strokes)[i].Points.size() == 1)
				{
					UndoList.push_back(undo_info);
					// update the undo/redo manager
					UndoRedoManager::checkIn(this);
					Strokes->erase(Strokes->begin() + i);
				}
				else
				{
					// else we search along the stroke...
					unsigned int j = 0;
					while(j < (*Strokes)[i].Points.size()-1)
					{
						math::Vec2f interval_begin = (*Strokes)[i].Points[j];
						math::Vec2f interval_end = (*Strokes)[i].Points[j+1];


						//we distinguish 4 cases:

						// ************** C A S E     A)  ************************************************************************************************** 
						// interval begin and interval end are outside the rubber circle
						if( !insideRubber(interval_begin, coords, rubber_radius) && !insideRubber(interval_end, coords, rubber_radius) )
						{
							// if interval begin and interval end are on opposite sides of the rubber circle we go on
							float interval_length = (interval_end - interval_begin).getLength();
							float dist_to_rubber1 = (coords - interval_begin).getLength();
							float dist_to_rubber2 = (coords - interval_end).getLength();

							if( abs(interval_length - dist_to_rubber1 - dist_to_rubber2) < 0.1f )
							{	
								// snap rubber onto stroke
								math::Vec2f snapped_coords = snapOntoStroke(interval_begin, interval_end, coords); 

								// we set two new points and generate a new stroke
								math::Vec2f new_stroke_end = getIntersectionPoint(interval_begin, snapped_coords, rubber_radius);
						
								Stroke new_stroke;
								new_stroke.Color = (*Strokes)[i].Color;
								new_stroke.ID = Brush2D::getStrokeID();
								new_stroke.LineWidth = (*Strokes)[i].LineWidth;
								//all the points before interval_begin will form a new stroke
								for(unsigned int k = 0; k <= j; k++)
								{
									new_stroke.Points.push_back((*Strokes)[i].Points[k]);
								}
								new_stroke.Points.push_back(new_stroke_end);

								// if the last stroke in the strokes list is empty we delete it
								if (Strokes->back().Points.size() == 0)
									Strokes->pop_back();

								// we push back the new stroke
								Strokes->insert(Strokes->begin()+i+1, new_stroke);
								// update undo information
								undo_info.NewStrokes.push_back(new_stroke);

								// we erase all the points from the old stroke and set the new start point
								if(j > 0)
								{
									(*Strokes)[i].Points.erase((*Strokes)[i].Points.begin(), (*Strokes)[i].Points.begin() + j);
								}
								(*Strokes)[i].Points.front() = getIntersectionPoint(interval_end, snapped_coords, rubber_radius);

								// since we cutted our new search start is the second element
								j = 1;
								continue;
							}
						}


						// ************** C A S E     B)  ************************************************************************************************** 
						// the interval start is outside the rubber and the interval end is inside
						if( !insideRubber(interval_begin, coords, rubber_radius) && insideRubber(interval_end, coords, rubber_radius) )
						{
							// snap rubber onto stroke
							math::Vec2f snapped_coords = snapOntoStroke(interval_begin, interval_end, coords); 

							// we set one new point and generate a new stroke
							math::Vec2f new_stroke_end = getIntersectionPoint(interval_begin, snapped_coords, rubber_radius);
						
							Stroke new_stroke;
							new_stroke.Color = (*Strokes)[i].Color;
							new_stroke.ID = Brush2D::getStrokeID();
							new_stroke.LineWidth = (*Strokes)[i].LineWidth;
							//all the points before interval_begin will form a new stroke
							for(unsigned int k = 0; k <= j; k++)
							{
								new_stroke.Points.push_back((*Strokes)[i].Points[k]);
							}
							new_stroke.Points.push_back(new_stroke_end);

							// if the last stroke in the strokes list is empty we delete it
							if (Strokes->back().Points.size() == 0)
								Strokes->pop_back();

							// we push back the new stroke
							Strokes->insert(Strokes->begin()+i+1, new_stroke);
							// update undo information
							undo_info.NewStrokes.push_back(new_stroke);

							// if the interval end is the last point from the old stroke we delete the old stroke
							if( (j + 1) == (*Strokes)[i].Points.size() - 1 )
							{
								Strokes->erase(Strokes->begin() + i);
								break;
							}
							else
							{
								// we erase all the j points from the old stroke
								(*Strokes)[i].Points.erase((*Strokes)[i].Points.begin(), (*Strokes)[i].Points.begin() + j + 1);
							}
						

							// our new search start is the first element of the trimmed old stroke ( = j+1 -th element of the untrimmed old stroke)
							j = 0;
							continue;
						}


						// ************** C A S E     C)  ************************************************************************************************** 
						// both interval start and interval end are inside the rubber shape
						if( insideRubber(interval_begin, coords, rubber_radius) && insideRubber(interval_end, coords, rubber_radius) )
						{
							// if the interval end is the last point from the old stroke we delete the stroke
							if( (j + 1) == (*Strokes)[i].Points.size() - 1 )
							{
								Strokes->erase(Strokes->begin() + i);
								break;
							}
							else
							{
								// we erase the first element
								(*Strokes)[i].Points.erase((*Strokes)[i].Points.begin());
							}
						
							// our new search start is the first element of the trimmed old stroke ( = j+1 -th element of the untrimmed old stroke)
							j = 0;
							continue;
						}


						// ************** C A S E     D)  ************************************************************************************************** 
						// the interval start is inside the rubber and the interval end is outside
						if( insideRubber(interval_begin, coords, rubber_radius) && !insideRubber(interval_end, coords, rubber_radius) )
						{
							// snap rubber onto stroke
							math::Vec2f snapped_coords = snapOntoStroke(interval_begin, interval_end, coords); 

							// we set the first point from the old stroke to a new position
							(*Strokes)[i].Points.front() = getIntersectionPoint(interval_end, snapped_coords, rubber_radius);

							// since we cutted our new search start is the second element
							j = 1;
							continue;
						}

						// nothing found: check the next interval
						j++;
					
					}

					// we set the trimmed stroke as a new one into the undo info
					undo_info.NewStrokes.push_back((*Strokes)[i]);
					UndoList.push_back(undo_info);
					// update the undo/redo manager
					UndoRedoManager::checkIn(this);
					// clear the redo list
					RedoList.clear();
					break;
				}
			}
		}
	}
}

void Rubber2D::setRubberSize(Pose& pose)
{
	RubberRadius += 50*(pose.LeftHandPosition.y - LastPose.LeftHandPosition.y);
	RubberRadius = max(RubberRadius, 1.0f);
}


bool Rubber2D::insideRubber(math::Vec2f& point, math::Vec2f& rubber_center, float rubber_radius)
{
	return ( (point - rubber_center).getLength() < rubber_radius );
}

math::Vec2f Rubber2D::getIntersectionPoint(math::Vec2f& point, math::Vec2f& rubber_center, float rubber_radius)
{
	float t = rubber_radius /  (point - rubber_center).getLength();
	return ( (1.0f - t) * rubber_center + t * point );
}

math::Vec2f Rubber2D::snapOntoStroke(math::Vec2f& point1, math::Vec2f& point2, math::Vec2f& rubber_center)
{
	math::Vec2f direction = point2 - point1;
	direction /= direction.getLength();
	math::Vec2f to_center = rubber_center - point1;
	float scale = direction.x * to_center.x + direction.y * to_center.y;

	return (point1 + scale * direction);
}

void Rubber2D::undo()
{
	if(UndoList.size() > 0)
	{
		// first we delete all new strokes...
		for(unsigned int n = 0; n < UndoList.back().NewStrokes.size(); n++)
		{
			for(unsigned int i = 0; i < Strokes->size(); i++)
			{
				if( (*Strokes)[i].ID == UndoList.back().NewStrokes[n].ID )
				{
					Strokes->erase(Strokes->begin() + i);
					break;
				}
			}
		}


		//...and repair the original stroke
		Strokes->push_back(UndoList.back().OldStroke);

		// push the undo information on the redo stack
		RedoList.push_back(UndoList.back());

		//pop the undo info from the undo stack
		UndoList.pop_back();
	}
}

void Rubber2D::redo()
{
	if(RedoList.size() > 0)
	{
		// we delete the original stroke from the stroke stack...
		for(unsigned int i = 0; i < Strokes->size(); i++)
		{
			if( (*Strokes)[i].ID == RedoList.back().OldStroke.ID )
			{
				Strokes->erase(Strokes->begin() + i);
				break;
			}
		}

		//...and push back the new strokes
		for(unsigned int i = 0; i < RedoList.back().NewStrokes.size(); i++)
		{
			Strokes->push_back(RedoList.back().NewStrokes[i]);
		}

		// push the info on the undo stack
		UndoList.push_back(RedoList.back());

		//pop the redo info from the redo stack
		RedoList.pop_back();
	}
}