#include "SmoothBrush.h"
#include "SOIL.h"


SmoothBrush::SmoothBrush(KinectPointer* right_hand, GridMesh* mesh)
{
	RightHandPointer = right_hand;
	BrushRadius = 4.0f;
	BrushStrength = 0.5f;
	Mesh = mesh;
	showBrush = false;

	TextureID = SOIL_load_OGL_texture("../images/smooth_brush.png", SOIL_LOAD_AUTO, SOIL_CREATE_NEW_ID, SOIL_FLAG_MIPMAPS);
}


SmoothBrush::~SmoothBrush(void)
{
}


void SmoothBrush::updateLeftHandInput(Pose &pose)
{
	setRadiusAndStrength(pose);
	LastPose = Pose(pose);
}

void SmoothBrush::updateRightHandInput(Pose &pose)
{
	processMesh();
	LastPose = Pose(pose);
}

void SmoothBrush::updateBothHandInput(Pose &pose)
{
	setRadiusAndStrength(pose);
	processMesh();
	LastPose = Pose(pose);
}

void SmoothBrush::update(Pose &pose)
{
	showBrush = (RightHandPointer->isActive())?(Mesh->ScreenToLocalCoords(RightHandPointer->getWindowCoordinates(), BrushCoordinates)):(false);
	LastPose = Pose(pose);
}

void SmoothBrush::render()
{
	// render brush
	if (showBrush)
	{
		GLUquadric* quadric = gluNewQuadric();
		glPushMatrix();
		glTranslatef(BrushCoordinates.x, BrushCoordinates.y + .1f, BrushCoordinates.z);

		// visualize strength by color and thickness of circle
		glColor3f(1.f, 1.0f - BrushStrength, 1.0f - BrushStrength);
		
		gluSphere(quadric,0.2,10,10);
		glRotatef(-90, 1.0f, 0.0f, 0.0f);
		
		gluDisk(quadric, BrushRadius - abs(BrushStrength)*0.5f - 0.2f, BrushRadius + abs(BrushStrength)*0.5f + 0.2f, 20, 1);

		gluDeleteQuadric(quadric);
		glPopMatrix();
	}
}

void SmoothBrush::setRadiusAndStrength(Pose& pose)
{
	float diff_x = pose.LeftHandPosition.x - LastPose.LeftHandPosition.x;
	float diff_y = pose.LeftHandPosition.y - LastPose.LeftHandPosition.y;

	if ( abs(diff_x) > abs(diff_y))
	{
		BrushRadius += 50 * diff_x;
		BrushRadius = max(0.0f, BrushRadius);
	}
	else 
	{
		BrushStrength += diff_y;
		BrushStrength = max(0.0f, BrushStrength);
		BrushStrength = min(1.0f, BrushStrength);
	}
}

void SmoothBrush::processMesh()
{
	showBrush = Mesh->ScreenToLocalCoords(RightHandPointer->getWindowCoordinates(), BrushCoordinates);
	// if brush is inside grid area
	if(showBrush)
	{
		
		float absradius = abs(BrushRadius);
		float sign = BrushRadius/absradius;

		int radius = (int) absradius;
		int brush_x = (int) BrushCoordinates.x;
		int brush_z = (int) BrushCoordinates.z;

		// we need a list for all the grid points that need to be processed
		std::vector<GridPoint*> points;
		std::vector<float> new_heights;


		// iterate over all the grid points in the radius area
		for (int x = brush_x -radius; x <= brush_x + radius; x++)
		{
			if ( (x >= 0) && (x <= Mesh->getWidth()) )
			{
				for (int z = brush_z - radius; z <= brush_z + radius; z++)
				{
					if ( (z >= 0) && (z <= Mesh->getLength()) )
					{
						GridPoint* p = Mesh->getGridPoint(x, z); 
						// we check if grid point is inside the radius
						float v1 = p->coords.x - BrushCoordinates.x;
						float v2 = p->coords.z - BrushCoordinates.z;
						float dist = sqrt(v1 * v1 + v2 * v2);
						if( dist < absradius )
						{
							// for every grid point we calculate the average height of it's neighbour points
							float average = 0;
							for(unsigned int i = 0; i < p->neighbours.size(); i++)
							{
								average += (Mesh->getGridPoint(p->neighbours[i]))->coords.y;
							}
							average /= p->neighbours.size();

							// calculate the new height of the grid point
							float height = (1 - BrushStrength) * p->coords.y + BrushStrength * average;
							// we cannot set the height immediately because this would affect the calculations of the neighbour points. hence we put the current point into a list
							points.push_back(p);
							new_heights.push_back(height);
						}
						
					}	
				}
			}
		}

		// for every point in the list we set the new heights
		for (unsigned int i = 0; i < points.size(); i++)
		{
			points[i]->coords.y = new_heights[i];
			// put the grid point into the normal recalculation list
			Mesh->recalculateGridPointNormal(points[i]);
		}
		
		// recalculate the involved grid point normals
		Mesh->executeNormalsRecalculation();
	}
}