#include "StraightLine2D.h"
#include "SOIL.h"


StraightLine2D::StraightLine2D(KinectPointer* right_hand, std::vector<Stroke>* strokes, Manipulator2D* manipulator)
{
	RightHandPointer = right_hand;
	Strokes = strokes;
	Manipulator = manipulator;
	LineWidth = 10.0f;
	StartPoint = math::Vec2f(0.0f, 0.0f);
	EndPoint = math::Vec2f(0.0f, 0.0f);
	Drawed = false;
	TextureID = SOIL_load_OGL_texture("../images/straightline.png", SOIL_LOAD_AUTO, SOIL_CREATE_NEW_ID, SOIL_FLAG_MIPMAPS);
}
StraightLine2D::~StraightLine2D()
{
}

void StraightLine2D::updateLeftHandInput(Pose &pose)
{
	setLineWidth(pose);
	checkStroke();
	// update pose
	LastPose = Pose(pose);
}


void StraightLine2D::updateRightHandInput(Pose &pose)
{
	setStrokePoint();
	// update pose
	LastPose = Pose(pose);
}	

void StraightLine2D::updateBothHandInput(Pose &pose)
{
	setLineWidth(pose);
	setStrokePoint();
	// update pose
	LastPose = Pose(pose);
}
	

void StraightLine2D::update(Pose &pose)
{
	checkStroke();
	LastPose = Pose(pose);
}	

void StraightLine2D::render()
{
	if(RightHandPointer->isActive())
	{
		math::Vec2f current_coords = RightHandPointer->getWindowCoordinates();

		//render line between start point and current position
		glColor3fv(Colorpicker::CurrentColor.v);
		glLineWidth(LineWidth);
		glBegin(GL_LINES);
		glVertex3f(StartPoint.x, StartPoint.y, .0f);
		glVertex3f(current_coords.x, current_coords.y, .0f);
		glEnd();

		// render start and current point
		glPushMatrix();
		glTranslatef(StartPoint.x, StartPoint.y, 0.0f);
		GLUquadric *quadric1 = gluNewQuadric();
		gluDisk(quadric1, 0.0, LineWidth/2.0f, 16, 1);			
		glPopMatrix();

		glPushMatrix();
		glTranslatef(current_coords.x, current_coords.y, 0.0f);
		GLUquadric *quadric2 = gluNewQuadric();
		gluDisk(quadric2, 0.0, LineWidth/2.0f, 16, 1);			
		glPopMatrix();

		// free memory
		gluDeleteQuadric(quadric1);
		gluDeleteQuadric(quadric2);
	}

}


void StraightLine2D::setStrokePoint()
{
	math::Vec2f world_coords = RightHandPointer->getWindowCoordinates();
	math::Vec2f local_coords = Manipulator->WorldToLocalCoords(world_coords);
	
	
	// we only set a new point if it has a specific minimum distance to the start point regarding the line width
	if( (world_coords - StartPoint).getLength() > LineWidth )
	{
		EndPoint = local_coords;
		Drawed = true;
	}
}

void StraightLine2D::setLineWidth(Pose& pose)
{
	LineWidth += 50*(pose.LeftHandPosition.y - LastPose.LeftHandPosition.y);
	LineWidth = max(LineWidth, 1.0f);
}

void StraightLine2D::checkStroke()
{
	if(Drawed)
	{
		//if we currently drawed we set the line
		Strokes->back().Points.push_back(Manipulator->WorldToLocalCoords(StartPoint));
		Strokes->back().Points.push_back(EndPoint);
		Strokes->back().Color = Colorpicker::CurrentColor;
		Strokes->back().LineWidth = LineWidth;
		//we set an undo info
		UndoList.push_back(Strokes->back());
		// update the undo/redo manager
		UndoRedoManager::checkIn(this);
		// clear the redo list
		RedoList.clear();

		Drawed = false;
	}
	
	//prepare a new stroke
	if( (Strokes->size() == 0) || (Strokes->back().Points.size() != 0) )
	{
		Stroke stroke;
		stroke.ID = Brush2D::getStrokeID();
		Strokes->push_back(stroke);
	}

	StartPoint = RightHandPointer->getWindowCoordinates();
	EndPoint = StartPoint;
}

void StraightLine2D::undo()
{
	if(UndoList.size() > 0)
	{
		//clear stroke with same id from the stroke stack
		for (unsigned int i = 0; i < Strokes->size(); i++)
		{
			if((*Strokes)[i].ID == UndoList.back().ID)
			{
				Strokes->erase(Strokes->begin() + i);
				break;
			}
		}

		// save the info in the redo list
		RedoList.push_back(UndoList.back());

		//pop the element from the undo stack
		UndoList.pop_back();
	}
}

void StraightLine2D::redo()
{
	if(RedoList.size() > 0)
	{
		//get the last element and push it back on the stroke stack
		Strokes->push_back(RedoList.back());
		// also push it on the undo stack
		UndoList.push_back(RedoList.back());
		//pop it from the redo stack
		RedoList.pop_back();
	}
}