#include "TerraBrush.h"
#include "SOIL.h"

#define PI (3.14159f)

TerraBrush::TerraBrush(KinectPointer* right_hand, GridMesh* mesh)
{
	RightHandPointer = right_hand;
	BrushRadius = 4.0f;
	BrushCurvature = 1.0f;
	Mesh = mesh;
	showBrush = false;

	TextureID = SOIL_load_OGL_texture("../images/terra_brush.png", SOIL_LOAD_AUTO, SOIL_CREATE_NEW_ID, SOIL_FLAG_MIPMAPS);
}



TerraBrush::~TerraBrush(void)
{
}


void TerraBrush::updateLeftHandInput(Pose &pose)
{
	setRadiusAndCurvature(pose);
	LastPose = Pose(pose);
}

void TerraBrush::updateRightHandInput(Pose &pose)
{
	processMesh();
	LastPose = Pose(pose);
}

void TerraBrush::updateBothHandInput(Pose &pose)
{
	setRadiusAndCurvature(pose);
	processMesh();
	LastPose = Pose(pose);
}

void TerraBrush::update(Pose &pose)
{
	showBrush = (RightHandPointer->isActive())?(Mesh->ScreenToLocalCoords(RightHandPointer->getWindowCoordinates(), BrushCoordinates)):(false);
	LastPose = Pose(pose);
}

void TerraBrush::render()
{
	// render brush
	if (showBrush)
	{
		GLUquadric* quadric = gluNewQuadric();
		glPushMatrix();
		glTranslatef(BrushCoordinates.x, BrushCoordinates.y + .1f, BrushCoordinates.z);

		float radius = abs(BrushRadius);

		// use different colors for hills/dips
		if ( radius/BrushRadius < 0.0f )
			glColor3f(0.f, 0.f, 1.f);
		else
			glColor3f(1.f, 0.f, 0.f);
		
		
		gluSphere(quadric,0.2,10,10);
		glRotatef(-90, 1.0f, 0.0f, 0.0f);
		
		// render curvature/shape of resulting hills/dips with iso lines 
		float quarter_amp = pow(0.333f, 1.0f/BrushCurvature) * radius;
		float half_amp = pow(0.5f, 1.0f/BrushCurvature) * radius;
		float three_quarter_amp = pow(0.666f, 1.0f/BrushCurvature) * radius;

		gluDisk(quadric, quarter_amp - 0.1f, quarter_amp + 0.1f, 20, 1);
		gluDisk(quadric, half_amp - 0.1f, half_amp + 0.1f, 20, 1);
		gluDisk(quadric, three_quarter_amp - 0.1f, three_quarter_amp + 0.1f, 20, 1);
		gluDisk(quadric, radius - 0.3f, radius + 0.3f, 20, 1);

		gluDeleteQuadric(quadric);
		glPopMatrix();
	}
}

void TerraBrush::setRadiusAndCurvature(Pose& pose)
{
	float diff_x = pose.LeftHandPosition.x - LastPose.LeftHandPosition.x;
	float diff_y = pose.LeftHandPosition.y - LastPose.LeftHandPosition.y;

	if ( abs(diff_x) > abs(diff_y))
	{
		BrushRadius += 50 * diff_x;
	}
	else 
	{
		BrushCurvature += 20 * diff_y;
		BrushCurvature = max(0.2f, BrushCurvature);
		BrushCurvature = min(8.0f, BrushCurvature);
	}
}

void TerraBrush::processMesh()
{
	showBrush = Mesh->ScreenToLocalCoords(RightHandPointer->getWindowCoordinates(), BrushCoordinates);
	// if brush is inside grid area
	if(showBrush)
	{
		float absradius = abs(BrushRadius);
		float sign = BrushRadius/absradius;

		int radius = (int) absradius;
		int brush_x = (int) BrushCoordinates.x;
		int brush_z = (int) BrushCoordinates.z;

		// iterate over all the grid points in the radius area
		for (int x = brush_x -radius; x <= brush_x + radius; x++)
		{
			if ( (x >= 0) && (x <= Mesh->getWidth()) )
			{
				for (int z = brush_z - radius; z <= brush_z + radius; z++)
				{
					if ( (z >= 0) && (z <= Mesh->getLength()) )
					{
						GridPoint* p = Mesh->getGridPoint(x, z); 
						// we check if grid point is inside the radius
						float v1 = p->coords.x - BrushCoordinates.x;
						float v2 = p->coords.z - BrushCoordinates.z;
						float dist = sqrt(v1 * v1 + v2 * v2);
						if( dist < absradius )
						{
							// calculate growth by using a sinusoidal function
							p->coords.y += sign * 0.1f * absradius / sqrtf(BrushCurvature) * ( cos( PI *  pow(dist/absradius,BrushCurvature) ) + 1); 
							Mesh->recalculateGridPointNormal(p);
						}
					}	
				}
			}
		}

		// recalculate the involved grid point normals
		Mesh->executeNormalsRecalculation();
	}
}