#pragma once
#include <ctime>
#include <algorithm>



namespace math {


/**
* High quality random number generator,
* which  is a little bit slower than typical random number generators
* but with a period  about 3.138 x 10^57.
*/
struct random
{
private:
	unsigned long long u,v,w;
	double storedval;
public:

	///standard constructor uses system time as random seed
	random() :v(4101842887655102017LL), w(1),storedval(0.0)
	{
		unsigned long long seed = time(NULL);
		u = seed^v;
		unsigned long long ull;
		uniform(ull);
		v = u; uniform(ull);
		w = v; uniform(ull);
	}

	///constructor initializes random generator with given seed
	random(unsigned long long seed) :v(4101842887655102017LL), w(1),storedval(0.0)
	{
		u = seed^v;
		unsigned long long ull;
		uniform(ull);
		v = u; uniform(ull);
		w = v; uniform(ull);
	}
	
	///generates 64bit pseudo random integer
	void uniform(unsigned long long& rv)
	{
		u = u*2862933555777941757LL + 7046029254386353087LL;
		v ^= v >> 17; v ^= v <<31; v ^= v >> 8;
		w = 4294957665U*(w & 0xffffffff) + (w >> 32);
		unsigned long long x = u ^(u << 21); x ^= x >>35; x^=x << 4;
		rv= (x+v)^w;
	}

	///generates a 32bit pseudo random integer
	void uniform(unsigned int& rv)
	{
		unsigned long long v;
		uniform(v);
		rv= (unsigned int) v;
	}

	///generates a pseudo random integer between min and max
	void uniform(unsigned min, unsigned max, unsigned& rv)
	{
		uniform(rv);
		rv =rv%(max-min+1)+min;
	}


	///generates a pseudo random double-precision floating point number uniformly distributed between 0 and 1
	void uniform(double& rv)
	{
		unsigned long long v;
		uniform(v);
		rv= 5.42101086242752217E-20 * v;
	}

	///generates a pseudo random double-precision floating point number uniformly distributed between min and max
	void uniform(const double min,const double max, double &rv)
	{
		double v;
		uniform(v);
		rv= (max-min)*v + min;
	}
	
	///generates a pseudo random single-precision floating point number uniformly distributed between 0 and 1
	void uniform(float & rv)
	{
		double v;
		uniform(v);
		rv =(float) v;
	}

	///generates a pseudo random single-precision floating point number uniformly distributed between 0 and 1
	void uniform(const float min,const float max, float& rv)
	{
		float f;
		uniform(f);
		rv= (max-min)*f + min;
	}

	



	///generates a normal deviate double-precision floating point value with mu = 0 and sigma = 1
	void normal(double& rv)
	{
		double v1,v2,rsq,fac;
		if(storedval == 0.)
		{
			do{
				uniform(v1);
				uniform(v2);

				v1 = 2.0 * v1-1.0;
				v2 = 2.0 * v2-1.0;
				rsq = v1*v1+v2*v2;
			}while(rsq >= 1.0 || rsq == 0.0);
			fac = sqrt(-2.0*log(rsq)/rsq);
			storedval = v1*fac;
			rv= v2*fac;

		}else
		{
			fac = storedval;
			storedval = 0.;
			rv= fac;
		}
	}

	///generates a normal deviate single-precision floating point value with mu = 0 and sigma = 1
	void normal(float& rv)
	{
		double d;
		normal(d);
		rv = (float)d;
	}

	///generates a normal deviate double-precision floating point value with mu and sigma 
	void normal(const double mu, const double sigma, double &rv)
	{	
		double d;
		normal(d);
		rv =mu + sigma * d;
	}

	///generates a normal deviate single-precision floating point value with mu and sigma 
	void normal(const float mu, const float sigma, float& rv)
	{	float f;
		normal(f);
		rv= mu + sigma * f;
	}	

};


}
